__author__ = 'walker'
"""
db_utils
Utils for working with backstage databases
"""
import os
import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from django.conf import settings
from django.core.management import call_command

def get_dsn(inst):
    """ build the psycopg2 connection dsn
    @param inst:
    @return:
    """
    try:
        if inst.dsn is not None and inst.dsn_backstage is not None:
            return inst.dsn, inst.dsn_backstage
    except:
        pass
    db = get_default_db(inst)
    #dsn for the default database
    dsn = "dbname=%s port=%s host=%s" % (db['NAME'], db['PORT'], db['HOST'])
    #dsn for the 'backstage' database
    dsn_backstage = "dbname=backstage port=%s host=%s" % (db['PORT'], db['HOST'])
    inst.dsn = dsn
    inst.dsn_backstage = dsn_backstage
    return dsn, dsn_backstage

def get_default_db(inst):
    """
    return the instance's default db
    @param inst:
    @return:
    """
    try:
        db = inst.settings.DATABASES['default']
        return db
    except:
        print 'error getting database'
        return None


def connect_default(inst):
    """
    Connect this instance to its default database, as defined in settings
    @param inst: Backstage Act or Venue instance
    @return:
    """
    try:
        dsn, dsn_backstage = get_dsn(inst)
        conn = psycopg2.connect(dsn)
        inst.conn = conn
        return conn
    except psycopg2.OperationalError, e:
        if 'does not exist' in str(e):
            s = str(e).replace('FATAL:', '').strip()
            s += '\nHint: try "backstage.db.db_utils.create_default()"'
            print s
            return None
        else:
            print e
            return

def sync_default(inst):

    """
    Sync the default database
    @param inst:
    @return:
    """
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', '%s.settings' % inst.name)
    os.environ['DJANGO_SETTINGS_MODULE'] = '%s.settings' % inst.name
    call_command('syncdb', verbosity=0, interactive=False, settings='%s.settings' % inst.name)

def migrate_default(inst):
    """
    Migrate the default database
    @param inst:
    @return:
    """
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', '%s.settings' % inst.name)
    os.environ['DJANGO_SETTINGS_MODULE'] = '%s.settings' % inst.name
    call_command('migrate', interactive=False, verbosity=0, settings='%s.settings' % inst.name)

def create_default(inst):
    """
    Create a backstage database for a given backstage instance
    @param inst:
    @return:
    """
    #first try to connect to it. fail if it already exists
    try:
        dsn, dsn_backstage = get_dsn(inst)
    except Exception, e:
        print 'Error getting the dsn info'
        print e
        raise
    try:
        conn = psycopg2.connect(dsn)
        if isinstance(conn, psycopg2._psycopg.connection):
            print 'Database already exists'
            return
    except: # failure to connect, ie we can continue
        pass
    #try to connect to the backstage database. necessary to
    #connect to an existing db before we can create this one
    try:
        conn = psycopg2.connect(dsn_backstage)
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cur = conn.cursor()
        db = get_default_db(inst)
        dbname = db['NAME']
        q = "CREATE DATABASE %s" % dbname
        cur.execute(q)
        conn.commit()
        cur.close()
        conn.close()
        print 'Successfully created Act database %s' % dbname
        return
    except Exception, e:
        print e
        raise



from .db_settings import DATABASES
from .common_settings import *
#Import all venue-specific settings
from .venue_settings import *

for app in VENUE_APPS:
    app in INSTALLED_APPS or INSTALLED_APPS.append(app)



import os
import sys
import subprocess
import stat

from backstage.utils import uwsgi_portsniffer
from backstage.utils.uwsgi.uwsgi_utils import build_uwsgi
from backstage.db.db_utils import create_default

class Venue():
    """A backstage Venue is a specific local install of backstage."""

    def __init__(self, venue_home):
        """Venue folder structure must already exist before we can create and utilize
        a venue instance
        """
        venue_secret = os.path.join(venue_home, '.LIVE/backstage-venue.txt')
        if not os.path.exists(venue_secret):
            print 'Invalid Venue'
            print 'Hint: create a Venue first using backstage.shortcuts.new_venue'
            return None
        #so now we have a valid venue folder structure we can instantiate the object
        self.venue_home = venue_home.rstrip('/') # or it will play with the auto pathing
        self.venue_root, self.venue_name = os.path.split(self.venue_home)
        self.name = self.venue_name
        self.longname = 'backstage-%s' % self.venue_name
        self.acts_root = os.path.abspath(os.path.join(self.venue_home, 'acts'))
        self.dbname = self.longname.replace('-', '_')

        if not self.venue_home in sys.path:
            sys.path.insert(0, self.venue_home)
        if not self.venue_root in sys.path:
            sys.path.insert(0, self.venue_root)
        if not self.acts_root in sys.path:
            sys.path.insert(2, self.acts_root)
        self.settings = None
        self.acts = {}
        self.conn = None
        self.get_settings()
        self.get_acts()
        self.uwsgi_config, self.uwsgi_ini = build_uwsgi(self, 'venue')
        try:
            self.connect()
        except:
            print 'OK, continuing, you can create the database later'
        return

    def get_settings(self):
        """ import the venue's settings.py"""
        settings = None
        try:
            exec_string = 'from %s import settings' % self.venue_name
            exec exec_string
            self.settings = settings
            return True
        except:
            print 'error getting venue settings'
            raise

    def connect(self):
        """
        connect to the instance's default database
        @return:
        """
        from backstage.db.db_utils import connect_default
        self.conn = connect_default(self)

    def create_all_act_dbs(self):
        for actname, actinstance in self.acts.iteritems():
            a = actinstance
            try:
                create_default(a)
            except:
                pass
        self.connect_all_act_dbs()

    def connect_all_act_dbs(self):
        """
        try to connect to all act databases
        @return:
        """
        act_connections = {}
        for k, v in self.acts.iteritems():
            name = k
            instance = v
            try:
                instance.connect()
                act_connections[name] = instance.conn.dsn
            except:
                act_connections[name] = None
        self.act_connections = act_connections
        return

    def dumpsettings(self):
        try:
            outfile = os.path.join(self.venue_home, 'settings_dump.py')
            of = open(outfile, 'w')
            my_settings = dir(self.settings)
            for my_setting in my_settings:
                s = "%s = my_settings['%s']\n" % (my_setting, my_setting)
                of.write(s)
            of.close()
        except:
            err = 'ERROR dumping settings'
            raise
        return True

    def build_virtualenv(self):
        """build the virtual environment for this backstage venue"""
        cwd = os.getcwd()
        venvdir = os.path.join(self.venue_home, 'venv')
        cmd = '%s/build_virtualenv' % (venvdir)
        st = os.stat(cmd)
        os.chmod(cmd, st.st_mode | stat.S_IEXEC)
        os.chdir(venvdir)
        subprocess.call(cmd)
        os.chdir(cwd)
        return

    def get_uwsgi_port(self):
        self.uwsgi_ip, self.uwsgi_port = uwsgi_portsniffer.get_uwsgi_port(self)
        return

    def reload(self):
        """Reload the Venue, by touching its ini file"""
        try:
            with file(self.uwsgi_ini, 'a'):
                os.utime(self.uwsgi_ini, None)
        except IOError:
            print 'Could not update, permission denied.'
            return
        self.get_uwsgi_port()
        return

    def get_acts(self):
        """get this Venue's Acts and find out about them"""
        from backstage.shortcuts import Act
        acts_list = os.listdir(self.acts_root)
        for a in acts_list:
            test_act = Act(self, a)
            try:
                if 'keyfile' in test_act.__dict__:
                    self.acts[a] = test_act
                else:
                    pass
            except:
                pass

    def __unicode__(self):
        s = 'Backstage Venue instance %s at %s' % (self.venue_name, self.venue_root)
        return s

    def start_all_acts(self):
        for act in self.acts.values():
            act.start()

    def stop_all_acts(self):
        for act in self.acts.values():
            act.stop()

    def restart_all_acts(self):
        for act in self.acts.values():
            act.restart()


import sys
import os
p = os.path.dirname(os.path.abspath(__file__))  # settings
ACT_HOME, n2 = os.path.split(p)
ACTS_DIR, ACT_NAME = os.path.split(ACT_HOME)
VENUE_PATH = os.path.dirname(ACTS_DIR)
VENUE_ROOT, VENUE_NAME = os.path.split(VENUE_PATH)
INSTRUMENTS_ROOT = os.path.abspath(os.path.join(VENUE_PATH, 'instruments'))

#prepend into sys.path in reverse order of importance. The Act itself should
#automatically already be present first in the list
INSTRUMENTS_ROOT in sys.path or sys.path.insert(1, INSTRUMENTS_ROOT)
VENUE_ROOT in sys.path or sys.path.insert(1, VENUE_ROOT)
VENUE_PATH in sys.path or sys.path.insert(1, VENUE_PATH)
ACTS_DIR in sys.path or sys.path.insert(1, ACTS_DIR)
syspath = sys.path
exec('from %s.settings import *' % VENUE_NAME)

from theme_settings import *

from db_settings import DATABASES
DATABASES['default']['NAME'] = 'backstage_%s_%s' % (VENUE_NAME, ACT_NAME)

from act_settings import *


TEMPLATE_DIRS.insert(0, ACT_NAME + '/templates')
TEMPLATE_DIRS.append('%s/instruments/local/templates' % VENUE_ROOT)
TEMPLATE_DIRS.append('%s/instruments/local/templates/ads' % VENUE_ROOT)

ROOT_URLCONF = '%s.urls' % (ACT_NAME)

for act in ACT_APPS:
    act in INSTALLED_APPS or INSTALLED_APPS.append(act)
INSTALLED_APPS.insert(0, ACT_NAME)

try:
    from act_settings import WWW_DEFAULT_HOSTS
except ImportError:
    WWW_DEFAULT_HOSTS = ['', 'www']
try:
    from act_settings import WWW_DEFAULT_TLD
except ImportError:
    WWW_DEFAULT_TLD = 'com'
try:
    from act_settings import WWW_SERVER_NAME_LIST
except ImportError:
    WWW_SERVER_NAME_LIST = []
if len(WWW_SERVER_NAME_LIST) == 0:
    for h in WWW_DEFAULT_HOSTS:
        server = '%s.%s.%s' % (h, ACT_NAME, WWW_DEFAULT_TLD)
        server = server.lstrip('.')
        WWW_SERVER_NAME_LIST.append(server)
WWW_SERVER_NAMES = ' '.join(WWW_SERVER_NAME_LIST)

del p,n2
sys.path = syspath


#act_settings.py
SITE_ID = 1
SITE_DB = 'None'
DEBUG = False
TEMPLATE_DEBUG = False

ACT_APPS = [] # APPS to include in INSTALLED_APPS

WWW_DEFAULT_TLD = '.com'
WWW_DEFAULT_HOSTS = ['', 'www']
WWW_SERVER_NAME_LIST = []
# List of web servers this act services
#if empty it will be generated from WWW_DEFAULT_TLD and WWW_DEFAULT_HOSTS


"""
linker_pg_plugin.py
Creates vassals for the uWSGI emperor using the pg_emperor plugin.
defines start, stop, and restart functions.  Upstream classes such as Act
can choose which linker from which to load these (eg file:// vs pg_emperor, etc)
"""

import time
from backstage.utils.uwsgi.uwsgi_utils import table, uid, gid

def start(inst):
    try:
        conn = inst.connect()
        cur = conn.cursor()
    except:
        print 'connection error'
        return
    try:
        ts = time.time()
        q = "select * from %s where name = '%s.ini'" % (table, inst.name)
        cur.execute(q)
        conn.commit()
        if cur.fetchone() != None:
            print '%s found in wsgi table.\nHint try restart() to force a re-start' % (inst.name)
            return
        q = "insert into %s (name,config,ts,uid,gid) values ('%s.ini','%s',%s,'%s','%s')" % \
            (table, inst.name, inst.uwsgi_ini, ts, uid, gid)
        cur.execute(q)
        conn.commit()
        print 'start request submitted for %s' % inst.name
    except:
        raise
        print 'failed to start'
        conn.rollback
    cur.close()
    conn.close()
    return


def stop(inst):
    try:
        conn = inst.connect()
        cur = conn.cursor()
    except:
        print 'connection error'
        return
    try:
        q = "SELECT * FROM %s WHERE name = '%s.ini'" % (table, inst.name)
        cur.execute(q)
        res = cur.fetchone()
        conn.commit()
        if res is None:
            print 'Act %s is not running' % inst.name
            return
    except:
        cur.close()
        conn.close()
        raise
    try:
        q = "DELETE FROM %s WHERE name = '%s.ini'" % (table, inst.name)
        cur.execute(q)
        conn.commit()
        cur.close()
        conn.close()
        print 'stop request submitted for %s' % inst.name
    except:
        print 'delete error %s' % inst.name
        conn.rollback()
        cur.close()
        conn.close()
    return


def restart(inst):
    try:
        conn = inst.connect()
        cur = conn.cursor()
    except:
        print 'connection error'
        return
    try:
        q = "SELECT * FROM %s WHERE name ='%s.ini'" % (table, inst.name)
        cur.execute(q)
        conn.commit()
        if cur.fetchone() == None:
            print '%s is not running. Attempting to start.' % inst.name
            start(inst)
            cur.close()
            conn.close()
            return
        ts = time.time()

        q = "UPDATE %s set ts = %s where name = '%s.ini'" % (table, ts, inst.name)
        cur.execute(q)
        conn.commit()
        cur.close()
        conn.close()
        print 'submitted re-start request for %s' % inst.name
    except:
        print 'failed %s' % inst.name
        conn.rollback
        cur.close()
        conn.close()


"""
settings_tags.py
This file provides a simple Django Templatetag
to expose the app's settings values in the HTML 
template environment. Settings are stored as 'csettings'
to avoid a name conflict 
and accesses as {{ csettings.FOO }} in the templates.
""" 

from django import template
from django.conf import settings

register = template.Library()

# settings value
@register.simple_tag
def csettings(name):
    """ provides for the inclusion of setting variables within a template. """
    return getattr(settings, name, "")



import os
"""
linker_file_ini.py
Creates vassals for the uWSGI emperor using the pg_emperor plugin.
defines start, stop, and restart functions.  Upstream classes such as Act
can choose which linker from which to load these (eg file:// vs pg_emperor, etc)
"""
def start(inst):
    """
    Start a backstage instance by *linking* it's INI file
    @param inst: Backstage Instance, such as an Act
    @return:
    """
    if not os.path.exists(inst.uwsgi_file):
        with open(inst.uwsgi_file, 'w') as f:
            f.write(inst.uwsgi_ini)

    if os.path.exists(inst.uwsgi_vassal):
        try:
            a = inst.name
        except NameError:
            a = ''
        print 'Vassal file already exists. Hint: try %s.restart() instead' % a
        return
    os.symlink(inst.uwsgi_file, inst.uwsgi_vassal)
    print 'start request submitted for %s' % inst.name
    #inst.get_uwsgi_port()
    return


def stop(inst):
    """
    Stop a backstage instance by *unlinking* it's INI file
    @param inst: Backstage Instance, such as an Act
    @return:
    """
    if not os.path.exists(inst.uwsgi_vassal):
        print 'vassal %s not running' % inst.name
        return
    try:
        os.unlink(inst.uwsgi_vassal)
        print 'stop request submitted for %s' % inst.name
    except:
        print 'error unlinking vassal'
        return
    inst.uwsgi_port = None
    return


def restart(inst):
    """
    Re-start a backstage instance by *touching* it's INI file
    @param inst: Backstage Instance, such as an Act
    @return:
    """
    try:
        with file(inst.uwsgi_file, 'a'):
            os.utime(inst.uwsgi_file, None)
    except IOError:
        print 'Could not update, permission denied.'
        return
    print 're-start request submitted for %s' % inst.name
    #inst.get_uwsgi_port()
    return


import os

def readtext(filepath):
    try:
        file = open(filepath,'r')
        lines = file.read()
        file.close()
        return lines
    except:
        return None

def readFromMedia(MEDIA_ROOT,filename):
    try:
        file = open(os.path.join(MEDIA_ROOT,filename))
        lines = file.read()
        file.close()
        return lines
    except:
        return None



__author__ = 'walker'
"""
nginx.py
backstage tools for building and deploying nginx servers
"""
import os
import sys
import backstage
from backstage.shortcuts import Act, Venue

confdir = os.path.join(os.path.dirname(os.path.abspath(backstage.__file__)), 'conf')
confsrc = os.path.join(confdir, 'nginx.conf.src')

def build_nginx_conf(instance):
    outdir = outpath(instance)
    outfile = instance.longname + '.conf'
    of = open(os.path.join(outdir, outfile), 'w')
    with open(confsrc, 'r') as infile:
        confdata = infile.read()

    # data to substitute into the content
    longname = instance.longname
    server_list_string = instance.settings.WWW_SERVER_NAMES
    listen_addr = instance.settings.NGINX_LISTEN_ADDR
    if isinstance(instance, Act):
        venue_path = instance.venue.venue_home
    elif isinstance(instance, Venue):
        venue_path = instance.venue_home
    confdata = confdata.format(longname=longname, venue_path=venue_path, server_list_string=server_list_string, listen_addr=listen_addr)
    of.write(confdata)
    of.close()
    return

def outpath(instance):
    op = ''
    if isinstance(instance, Act):
        op = instance.venue.venue_home
    elif isinstance(instance, Venue):
        op = instance.venue_home
    else:
        print 'instance error'
        raise
    op += '/.LIVE'
    return op